import pytest
from xchainpy_client.models import tx_types
from xchainpy_client.models.types import Network
from xchainpy_bitcoincash.models.client_types import BitcoincashClientParams, BitcoincashTxParams
from xchainpy_bitcoincash.client import Client
from xchainpy_util.asset import AssetBCH


class TestBitcoincashClient:
    memo = 'SWAP:THOR.RUNE'
    phrase = 'atom green various power must another rent imitate gadget creek fat then'
    testnet_address_path0 = 'bchtest:qpd7jmj0hltgxux06v9d9u6933vq7zd0kyjlapya0g'
    testnet_address_path1 = 'bchtest:qrkd7dhu7zcmn6wwvj3p4aueslycqchj5vxx3stmjz'
    mainnet_address_path0 = 'bitcoincash:qp4kjpk684c3d9qjk5a37vl2xn86wxl0f5j2ru0daj'
    mainnet_address_path1 = 'bitcoincash:qr4jrkhu3usuk8ghv60m7pg9eywuc79yqvd0wxt2lm'
    
    @pytest.fixture
    def client(self):
        self.client = Client(BitcoincashClientParams(network='testnet', phrase=self.phrase))
        yield
        self.client.purge_client()

    def test_set_phrase_should_return_correct_address(self, client):
        self.client.set_network('testnet')
        assert self.client.set_phrase(self.phrase) == self.testnet_address_path0

        self.client.set_network('mainnet')
        assert self.client.set_phrase(self.phrase) == self.mainnet_address_path0

    def test_set_phrase_with_derivation_path_should_return_correct_address(self, client):
        self.client.set_network('testnet')
        self.client.set_phrase(self.phrase)
        self.client.get_address(index=1) == self.testnet_address_path1

        self.client.set_network('mainnet')
        self.client.set_phrase(self.phrase)
        self.client.get_address(index=1) == self.mainnet_address_path1

    def test_invalid_phrase(self):
        with pytest.raises(Exception) as err:
            assert Client(BitcoincashClientParams(phrase='invalid phrase'))
        assert str(err.value) == "invalid phrase"

    def test_right_phrase(self, client):
        assert self.client.set_phrase(self.phrase) == self.testnet_address_path0

    def test_validate_address(self, client):
        assert self.client.validate_address(address=self.mainnet_address_path0)
        assert self.client.validate_address(address=self.testnet_address_path0)

    def test_should_return_valid_explorer_url(self, client):
        assert self.client.get_explorer_url() == 'https://www.blockchain.com/bch-testnet'
        self.client.set_network(Network.Mainnet)
        assert self.client.get_explorer_url() == 'https://www.blockchain.com/bch'

    def test_should_retrun_valid_explorer_address_url(self, client):
        assert self.client.get_explorer_address_url('testAddressHere') == 'https://www.blockchain.com/bch-testnet/address/testAddressHere'
        self.client.set_network(Network.Mainnet)
        assert self.client.get_explorer_address_url('anotherTestAddressHere') == 'https://www.blockchain.com/bch/address/anotherTestAddressHere'

    def test_should_retrun_valid_explorer_tx_url(self, client):
        assert self.client.get_explorer_tx_url('testTxHere') == 'https://www.blockchain.com/bch-testnet/tx/testTxHere'
        self.client.set_network(Network.Mainnet)
        assert self.client.get_explorer_tx_url('anotherTestTxHere') == 'https://www.blockchain.com/bch/tx/anotherTestTxHere'

    @pytest.mark.asyncio
    async def test_has_balances(self, client):
        balance = await self.client.get_balance('bchtest:qz35h5mfa8w2pqma2jq06lp7dnv5fxkp2svtllzmlf')
        balance = balance[0]
        assert balance.asset == AssetBCH
        assert isinstance(balance.amount, float)

    @pytest.mark.asyncio
    async def test_get_transaction_with_hash(self, client):
        tx_data = await self.client.get_transaction_data('0d5764c89d3fbf8bea9b329ad5e0ddb6047e72313c0f7b54dcb14f4d242da64b')
        assert tx_data.tx_hash == '0d5764c89d3fbf8bea9b329ad5e0ddb6047e72313c0f7b54dcb14f4d242da64b'
        assert len(tx_data.tx_from) == 1
        assert tx_data.tx_from[0].address == 'bchtest:qzyrvsm6z4ucrhaq4zza3wylre7mavvldgr67jrxt4'
        assert str(tx_data.tx_from[0].amount) == '0.04008203'

        assert len(tx_data.tx_to) == 2
        assert tx_data.tx_to[0].address == 'bchtest:qq235k7k9y5cwf3s2vfpxwgu8c5497sxnsdnxv6upc'
        assert str(tx_data.tx_to[0].amount) == '0.04005704'

    @pytest.mark.asyncio
    async def test_get_transactions(self, client):
        tx_page = await self.client.get_transactions(tx_types.TxHistoryParams(address='bchtest:qz35h5mfa8w2pqma2jq06lp7dnv5fxkp2svtllzmlf', limit=1))
        tx_data = tx_page.txs[0]
        assert tx_data.tx_hash == '0957b51a39d6e67a7a3ced07b49a1102006cb51cea7c82b5a949a8678f3ac35c'
        assert len(tx_data.tx_from) == 1
        assert tx_data.tx_from[0].address == 'bchtest:qzmpc0fz8tdz9kkfhxzmu0rt6d23dvyusugshegndx'
        assert tx_data.tx_from[0].amount == 0.00003834

        assert len(tx_data.tx_to) == 2
        assert tx_data.tx_to[0].address == 'bchtest:qz35h5mfa8w2pqma2jq06lp7dnv5fxkp2svtllzmlf'
        assert tx_data.tx_to[0].amount == 0.000006


    @pytest.mark.asyncio
    async def test_transfer_with_memo_and_fee_rate(self, client):
        fee_rates = await self.client.get_fee_rates()
        fee_rate = fee_rates.fast
        balance = await self.client.get_balance('bchtest:qzt6sz836wdwscld0pgq2prcpck2pssmwge9q87pe9')
        balance = balance[0]
        if balance.amount > 0:
            amount = 0.00000001
            tx_id = await self.client.transfer(BitcoincashTxParams(amount, 'bchtest:qzt6sz836wdwscld0pgq2prcpck2pssmwge9q87pe9', self.memo, fee_rate))
            assert tx_id
            
    @pytest.mark.asyncio
    async def test_fee_and_rates_normal_tx(self, client):
        fees_and_rates = await self.client.get_fees_with_rates()
        fees = fees_and_rates.fees
        rates = fees_and_rates.rates
        assert fees.fastest
        assert fees.fast
        assert fees.average
        assert rates.fastest
        assert rates.fast
        assert rates.average

    @pytest.mark.asyncio
    async def test_fee_and_rates_with_memo(self, client):
        fees_and_rates = await self.client.get_fees_with_rates(self.memo)
        fees = fees_and_rates.fees
        rates = fees_and_rates.rates
        assert fees.fastest
        assert fees.fast
        assert fees.average
        assert rates.fastest
        assert rates.fast
        assert rates.average

    @pytest.mark.asyncio
    async def test_estimated_fees_normal_tx(self, client):
        fees = await self.client.get_fees()
        assert fees.fastest
        assert fees.fast
        assert fees.average

    @pytest.mark.asyncio
    async def test_different_fees_normal_tx(self, client):
        fees = await self.client.get_fees()

        assert fees.fastest > fees.fast
        assert fees.fast > fees.average


import pytest
from xchainpy_bitcoincash import utils

class TestBitcoinCashUtils:
    memo = 'SWAP:THOR.RUNE'
    testnet_address = 'bchtest:qpd7jmj0hltgxux06v9d9u6933vq7zd0kyjlapya0g'
    mainnet_address = 'bitcoincash:qp4kjpk684c3d9qjk5a37vl2xn86wxl0f5j2ru0daj'


    def test_get_right_vault_fee(self):
        assert utils.calc_fee(fee_rate=10, memo=self.memo) == 1030

    def test_get_normal_fee(self):
        assert utils.calc_fee(fee_rate=10) == 780


from typing import List
from . import utils
from .models.api_types import AddressBalance, Block, Transaction, TransactionInput, TransactionOutput, TxUnspent
import http3
import json

def to_haskoin_network(network: str):
    return 'bchtest' if network == 'testnet' else 'bch'

async def get_account(haskoin_url:str, network:str, address:str) -> AddressBalance:
    """Get account from address
    https://api.haskoin.com/#/Address/getBalance

    :param client_url: The haskoin API url
    :type client_url: str
    :param network: network
    :type network: str
    :param address: The BCH address
    :type address: str
    :returns: AddressBalance
    """
    try:
        api_url = f'{haskoin_url}/{to_haskoin_network(network)}/address/{address}/balance'

        client = http3.AsyncClient(timeout=5)
        response = await client.get(api_url)

        if response.status_code == 200:
            balance_response = json.loads(response.content.decode('utf-8'))
            if "error" in balance_response:
                raise Exception(
                    f'Error is : {balance_response["error"]}\nMessage is : {balance_response["message"]}')
            else:
                result = AddressBalance(balance_response['received'], balance_response['utxo'], balance_response['address'],
                                        balance_response['txs'], balance_response['unconfirmed'], balance_response['confirmed'])
                return result
        else:
            return None
    except Exception as err:
        raise Exception(str(err))

async def get_transactions(haskoin_url:str, network:str, address:str, offset:int, limit:int):
    """Get address information
    https://api.haskoin.com/#/Address/getAddressTxsFull

    :param haskoin_url: haskoin url
    :type haskoin_url: str
    :param network: mainnet or testnet
    :type network: str
    :param address: wallet address
    :type address: str
    :returns: The fees with memo
    """
    try:
        api_url = f'{haskoin_url}/{to_haskoin_network(network)}/address/{address}/transactions/full'
        api_url += f'?offset={offset}&limit={limit}'

        client = http3.AsyncClient(timeout=5)
        response = await client.get(api_url)

        if response.status_code == 200:
            return json.loads(response.content.decode('utf-8'))
        else:
            raise Exception('failed to query transactions')
    except Exception as err:
        raise Exception(str(err))


async def get_transaction(haskoin_url:str, network:str, tx_id:str):
    """Get transaction by hash
    https://api.haskoin.com/#/Transaction/getTransaction

    :param client_url: The haskoin API url
    :type client_url: str
    :param network: network
    :type network: str
    :param tx_id: The transaction id
    :type tx_id: str
    :returns: Transaction info
    :raises: 'failed to query transaction by a given hash' if failed to query transaction by a given hash
    """
    try:
        api_url = f'{haskoin_url}/{to_haskoin_network(network)}/transaction/{tx_id}'

        client = http3.AsyncClient(timeout=5)
        response = await client.get(api_url)

        if response.status_code == 200:
            result = json.loads(response.content.decode('utf-8'))
            return result
        else:
            raise Exception('failed to query transaction by a given hash')
    except Exception as err:
        raise Exception(str(err))


async def get_suggested_tx_fee():
    """Get suggested fee amount for Bitcoin cash. (fee per byte)

    Note: Haskcoin does not provide fee rate related data
    So use Bitgo API for fee estimation
    Refer: https://app.bitgo.com/docs/#operation/v2.tx.getfeeestimate

    :returns: The Bitcoin cash stats
    """
    try:
        api_url = 'https://app.bitgo.com/api/v2/bch/tx/fee'

        client = http3.AsyncClient(timeout=5)
        response = await client.get(api_url)

        if response.status_code == 200:
            fee_response = json.loads(response.content.decode('utf-8'))
            return fee_response["feePerKb"] / 1000 # feePerKb to feePerByte
        else:
            fee_response = json.loads(response.content.decode('utf-8'))
            if "error" in fee_response:
                raise Exception(f'Error is : {fee_response["error"]}')

            return utils.DEFAULT_SUGGESTED_TRANSACTION_FEE
    except:
        return utils.DEFAULT_SUGGESTED_TRANSACTION_FEE

async def get_unspent_transactions(haskoin_url, network, address):
    """Get unspent transactions

    :param haskoin_url: haskoin url
    :type haskoin_url: str
    :param network: network
    :type network: str
    :param address: The BCH address
    :type address: str
    :returns: The Bitcoin cash stats
    :raises: 'failed to query unspent transactions' if failed to query unspent transactions
    """
    try:
        account = await get_account(haskoin_url, network, address)

        api_url = f'{haskoin_url}/{to_haskoin_network(network)}/address/{address}/unspent?limit={account.txs}'

        client = http3.AsyncClient(timeout=5)
        response = await client.get(api_url)

        if response.status_code == 200:
            result = json.loads(response.content.decode('utf-8'))
            return result
        else:
            raise Exception('failed to query unspent transactions')
    except Exception as err:
      raise Exception(str(err))

# async def broadcast_tx(client_url, tx_hex):
#     """Broadcast transaction
#     https://sochain.com/api#send-transaction

#     :param client_url: The haskoin API url
#     :type client_url: str
#     :param tx_hex: tranaction hex
#     :type tx_hex: str
#     :returns: Transaction ID
#     """
#     try:
#         api_url = f'{client_url}/transactions'

#         client = http3.AsyncClient(timeout=5)
#         response = await client.post(url=api_url, data=tx_hex)

#         if response.status_code == 200:
#             res = json.loads(response.content.decode('utf-8'))['data']
#             return res['txid']
#         else:
#             return json.loads(response.content.decode('utf-8'))['data']
#     except Exception as err:
#         raise Exception(str(err))


from xchainpy_client.models.tx_types import TxParams
from xchainpy_client.models.types import Network, RootDerivationPaths, XChainClientParams
from xchainpy_util.asset import Asset, AssetBCH


class BitcoincashClientParams(XChainClientParams):

    def __init__(self, network:Network=Network.Testnet, phrase=None,
                 root_derivation_paths:RootDerivationPaths=RootDerivationPaths(
                     mainnet="m/44'/145'/0'/0/",
                     testnet="m/44'/1'/0'/0/"),
                 haskoin_url:str='https://api.haskoin.com'):
        """
        :param network: network
        :type network: str
        :param phrase: phrase 
        :type phrase: str
        :param root_derivation_paths: root_derivation_paths 
        :type root_derivation_paths: RootDerivationPaths
        :param haskoin_url: haskoinUrl 
        :type haskoin_url: str
        """
        super().__init__(network, phrase, root_derivation_paths)
        self._haskoin_url = haskoin_url

    @property
    def haskoin_url(self):
        return self._haskoin_url

    @haskoin_url.setter
    def haskoin_url(self, haskoin_url):
        self._haskoin_url = haskoin_url

    @property
    def blockstream_url(self):
        return self._blockstream_url

    @blockstream_url.setter
    def blockstream_url(self, blockstream_url):
        self._blockstream_url = blockstream_url


class BitcoincashTxParams(TxParams):
    def __init__(self, amount, recipient, memo='', fee_rate=None, wallet_index=0, asset:Asset=AssetBCH):
        """Transfer BCH

        :param amount: amount of BCH to transfer (don't multiply by 10**8)
        :type amount: int, float, decimal
        :param recipient: destination address
        :type recipient: str
        :param memo: optional memo for transaction
        :type memo: str
        :param fee_rate: fee rates for transaction
        :type fee_rate: int
        :param wallet_index: wallet_index
        :type wallet_index: int
        """
        TxParams.__init__(self, AssetBCH, amount, recipient, memo, wallet_index)
        self._fee_rate = fee_rate

    @property
    def fee_rate(self):
        return self._fee_rate

    @fee_rate.setter
    def fee_rate(self, fee_rate):
        self._fee_rate = fee_rate





import json
import http3

BITCORE_URL = 'https://api.bitcore.io/api/BCH'

async def broadcast_tx(network, tx_hex):
    """Broadcast transaction

    :param network: testnet or mainnet
    :type network: str
    :param tx_hex: tranaction hex
    :type tx_hex: str
    :returns: Transaction ID
    """
    try:
        api_url = f'{BITCORE_URL}/{network}/tx/send'


        client = http3.AsyncClient(timeout=5)
        data = {
            'rawTx': tx_hex,
            'network': network,
            'coin': 'BCH'            
        }
        response = await client.post(url=api_url, json=data)

        if response.status_code == 200:
            res = json.loads(response.content.decode('utf-8'))['txid']
            return res
        else:
            raise Exception(response.content.decode('utf-8'))
    except Exception as err:
        raise Exception(str(err))